// librerie per RTC DS1307
#include <Wire.h>
#include <RTClib.h>

// librerie per tastiera
#include <Key.h>
#include <Keypad.h>

// librerie per la generazione dei codici OTP
#include <sha1.h>
#include <TOTP.h>

// seriale software per il display
#include <SoftwareSerial.h>

// config file
#include "CONFIG.h"


// variabili
RTC_DS1307 rtc;
Keypad keypad = Keypad(makeKeymap(keys), rowPins, colPins, KEY_ROWS, KEY_COLUMNS);
SoftwareSerial displaySerial(A1, A2);
TOTP totp = TOTP(hmacKey, 10);
char codiceInserito[7];
int codiceIndex;
char bufferSeriale[14];
int serialIndex;


void setup() {

  // inizializzazione RTC, viene configurato l'orario in cui lo sketch è stato compilato
  Wire.begin();
  rtc.begin();
  rtc.adjust(DateTime(__DATE__, __TIME__)); 

  // inizializzazione comunicazione seriale
  Serial.begin(9600);
  displaySerial.begin(9600);
  
  // attendo il boot del display e la scomparsa dello splash screen
  delay(500);

  Serial.println("Serratura OTP");
  Serial.println();

  // inizializzazione PIN
  pinMode(PIN_BUZZER, OUTPUT);
  pinMode(PIN_LEDROSSO, OUTPUT);
  pinMode(PIN_LEDVERDE, OUTPUT);
  pinMode(PIN_RELAY, OUTPUT);

  // inizializzazione variabili
  codiceIndex = -1;
  serialIndex = -1;

  // stato iniziale, serratura CHIUSA
  pulisciSchermo();
  cambiaStatoSerratura(false);
  aggiornaSecondaRiga();  
}

void loop() {

  // verifico se è stato ricevuto un carattere via seriale
  if (Serial.available() > 0) {

    byte inByte = Serial.read();

    // se il byte ricevuto è il carattere di NewLine (0x0A) interpreto il comando ricevuto
    if(inByte == 0x0A) eseguiComandoSeriale();

    // altrimenti lo memorizzo nel buffer
    else {
      serialIndex++;
      if(serialIndex == 14) serialIndex = 0;
      bufferSeriale[serialIndex] = inByte;
    }
  }

  // verifica se è stato premuto un tasto sulla tastiera
  char key = keypad.getKey();
  if (key != NO_KEY){

    // emetto un suono
    #ifdef BUZZER_ON
      tone(PIN_BUZZER, 440);
      delay(100);
      noTone(PIN_BUZZER);
    #endif 
    
    // se il tasto è un numero, lo visualizzo e memorizzo nell'array 
    if(key - '0' >= 0 && key - '0' < 10) {

      codiceIndex++;
      codiceInserito[codiceIndex] = key;
      displaySerial.write(key);

      // se sono stati inserite 6 cifre, verifico il codice
      if(codiceIndex == 5) verificaCodice();
    }

    // se il tasto è '*' cancello l'ultimo numero inserito
    else if(key == '*' && codiceIndex >= 0) {

      codiceIndex--;
      aggiornaSecondaRiga();
    }

    // se il tasto è '#' resetto l'inserimento
    else if(key == '#') {

      codiceIndex = -1;
      aggiornaSecondaRiga();    
    }
  }
}


void eseguiComandoSeriale() {

  // i comandi possibili sono:
  // ? visualizza data e ora letti dall'RTC
  // !ggMMaahhmmss imposta la nuova data e ora (gg/MM/aa hh:mm:ss) nell'RTC

  // resetto il buffer
  serialIndex = -1;
  
  if(bufferSeriale[0] == '?') stampaDataOra();
  
  else if(bufferSeriale[0] == '!') {

    // verifico che i dodici caratteri seguenti siano numeri
    for(int i = 1; i < 13; i++) 
      if(!isDigit(bufferSeriale[i])) {
        Serial.print(F("Carattere non valido (non e' un numero): "));
        Serial.println(bufferSeriale[i]);
        return;
      }

    // converto a due a due i caratteri in numero
    int giorno = (bufferSeriale[1] - '0') * 10 + (bufferSeriale[2] - '0');
    int mese = (bufferSeriale[3] - '0') * 10 + (bufferSeriale[4] - '0');
    int anno = 2000 + (bufferSeriale[5] - '0') * 10 + (bufferSeriale[6] - '0');
    int ora = (bufferSeriale[7] - '0') * 10 + (bufferSeriale[8] - '0');
    int minuti = (bufferSeriale[9] - '0') * 10 + (bufferSeriale[10] - '0');
    int secondi = (bufferSeriale[11] - '0') * 10 + (bufferSeriale[12] - '0');
    
    // eseguo controlli di validità dei dati
    if(giorno < 1 || giorno > 31) {
      Serial.print(F("Il giorno specificato non e' valido: "));
      Serial.println(giorno);
      return;      
    }
    if(mese < 1 || mese > 12) {
      Serial.print(F("Il mese specificato non e' valido: "));
      Serial.println(mese);
      return;      
    }
    if(ora > 24) {
      Serial.print(F("L'ora specificata non e' valida: "));
      Serial.println(ora);
      return;      
    }
    if(minuti > 59) {
      Serial.print(F("I minuti specificati nono sono validi: "));
      Serial.println(minuti);
      return;      
    }
    if(secondi > 59) {
      Serial.print(F("I secondi specificati non sono validi: "));
      Serial.println(secondi);
      return;      
    }
    
    // salvo l'orario e lo visualizzo aggiornato
    rtc.adjust(DateTime(anno, mese, giorno, ora, minuti, secondi));
    Serial.println(F("RTC aggiornato"));
    stampaDataOra();
  }

  else {
    Serial.print(F("Comando sconosciuto: "));
    Serial.println(bufferSeriale[0]);
  }
}

void stampaDataOra() {

  DateTime now = rtc.now();
  Serial.print(F("Data e ora attuali: "));
  Serial.print(now.day(), DEC); 
  Serial.print('/');
  Serial.print(now.month(), DEC);
  Serial.print('/');
  Serial.print(now.year(), DEC);
  Serial.print(" ");
  Serial.print(now.hour(), DEC);
  Serial.print(':');
  Serial.print(now.minute(), DEC);
  Serial.print(':');
  Serial.print(now.second(), DEC);
  Serial.println();
}


void cambiaStatoSerratura(bool serraturaAperta) {

  if(serraturaAperta) {
    digitalWrite(PIN_RELAY, HIGH);
    digitalWrite(PIN_LEDVERDE, HIGH);
    digitalWrite(PIN_LEDROSSO, LOW);
    displaySerial.write(254);
    displaySerial.write(128);
    displaySerial.write("     APERTA     ");
  }
  else {
    digitalWrite(PIN_RELAY, LOW);
    digitalWrite(PIN_LEDVERDE, LOW);
    digitalWrite(PIN_LEDROSSO, HIGH);
    displaySerial.write(254);
    displaySerial.write(128);
    displaySerial.write("     CHIUSA     ");         
  }
}


void pulisciSchermo() {
  
  displaySerial.write(254);
  displaySerial.write(128);
  displaySerial.write("                ");
  displaySerial.write("                ");
}


void aggiornaSecondaRiga() {

    displaySerial.write(254);
    displaySerial.write(192);
    displaySerial.write("PIN -->         ");
    displaySerial.write(254);
    displaySerial.write(200);
    for(int i = 0; i <= codiceIndex; i++)
      displaySerial.write(codiceInserito[i]);   
}


void verificaCodice() {

  // ottengo dal RTC il timestamp attuale e lo divido per 30 (timeStep di Google Authenticator)
  DateTime now = rtc.now();
  long GMT = now.unixtime();
  long timeStep = GMT / 30;

  // verifico se il codice digitato corrisponde ad uno dei 3 codici "vicini" al timestamp attuale
  bool codiceOk = false;
  for(int i = -1; i <= 1; i++) {  
    char* codiceGenerato = totp.getCodeFromSteps(timeStep + i);
    Serial.println(codiceGenerato);
    if(strcmp(codiceInserito, codiceGenerato) == 0) codiceOk = true;
  }

  // pulisco lo schermo
  pulisciSchermo();

  // se il codice è ok, emetto un suono e apro la serratura
  if(codiceOk) {

      #ifdef BUZZER_ON 
        tone(PIN_BUZZER, 660);
        delay(100);
        noTone(PIN_BUZZER);
        delay(200);
        tone(PIN_BUZZER, 660);
        delay(100);
        noTone(PIN_BUZZER); 
      #endif
    
    cambiaStatoSerratura(true);

    // se la modalità è MODO_IMPULSO, attendo TEMPO_APERTURA millisecondi quindi chiudo la serratura
    if(modoSerratura == MODO_IMPULSO) {
      delay(TEMPO_APERTURA);
      cambiaStatoSerratura(false);
    }

    // se invece è MODO_STABILE, attendo la pressione del tasto TASTO_CHIUSURA quindi chiudo la serratura
    else {
      char key;
      do {
        key = keypad.getKey();
      } while(key != TASTO_CHIUSURA);
      cambiaStatoSerratura(false);
    }
  }

  // se il codice è errato, visualizzo un messaggio e emetto un suono
  else {
  
    for(int i = 0; i < 3; i++) {
      
      #ifdef BUZZER_ON 
        tone(PIN_BUZZER, 220); 
      #endif
      
      displaySerial.write(254);
      displaySerial.write(128);
      displaySerial.write(" CODICE ERRATO! ");
      delay(500);
      
      #ifdef BUZZER_ON
        noTone(PIN_BUZZER); 
      #endif
      
      pulisciSchermo();
      delay(500);
    }
    displaySerial.write(254);
    displaySerial.write(128);
    displaySerial.write("     CHIUSA     ");    
  }

  // resetto 
  codiceIndex = -1;
  aggiornaSecondaRiga();
}

